import math
from .settings import *
from shapely import geometry

default_bounds = {'minlat':-90.0, 'minlon':-180.0, 'maxlat':90.0, 'maxlon':180.0}


def getDistanceFromCoord(lon1, lat1, lon2, lat2):
    # return km
    lon1, lat1, lon2, lat2 = map(math.radians, [lon1, lat1, lon2, lat2])
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = math.sin(dlat / 2) ** 2 + math.cos(lat1) * math.cos(lat2) * math.sin(dlon / 2) ** 2
    c = 2 * math.asin(math.sqrt(a))
    r = 6371
    return c * r


def getLineFromRefNodes(ref_node_list):
    if len(ref_node_list) < 2: return None, None
    point_list = [node.geometry for node in ref_node_list]
    line = geometry.LineString(point_list)
    point_list_xy = [node.geometry_xy for node in ref_node_list]
    line_xy = geometry.LineString(point_list_xy)
    return line, line_xy


def getPolygonFromRefNodes(ref_node_list):
    if len(ref_node_list) < 3: return None, None
    point_list = [node.geometry for node in ref_node_list]
    poly = geometry.Polygon(point_list)
    point_list_xy = [node.geometry_xy for node in ref_node_list]
    poly_xy = geometry.Polygon(point_list_xy)
    return poly, poly_xy


def printlog(msg,log_level='info'):
    if not print_log: return

    logger = logging.getLogger()

    if log_level == 'debug':
        logger.debug(msg)
    elif log_level == 'info':
        logger.info(msg)
    elif log_level == 'warning':
        logger.warning(msg)
    elif log_level == 'error':
        logger.error(msg)
    elif log_level == 'critical':
        logger.critical(msg)





from .classes import *
from shapely import ops


def _checkTpologyForTwoDegreeNodes(node):
    ib_link, ob_link = node.incoming_link_list[0], node.outgoing_link_list[0]
    # if not (ib_link.valid and ob_link.valid): return False
    if ib_link.from_node is ob_link.to_node: return False
    ib_link.ob_comb_link = ob_link
    return True


def _checkTpologyForFourDegreeNodes(node):
    ob_link_set = set()
    for ib_link in node.incoming_link_list:
        for ob_link in node.outgoing_link_list:
            if ib_link.from_node is not ob_link.to_node:
                if (ib_link.ob_comb_link is None) and (ob_link not in ob_link_set):
                    ib_link.ob_comb_link = ob_link
                    ob_link_set.add(ob_link)
                else:
                    return False
    for ib_link in node.incoming_link_list:
        if ib_link.ob_comb_link is None: return False
    return True


def _getNameForTheCombinedLink(ib_link):
    if ib_link.name == ib_link.ob_comb_link.name:
        return ib_link.name
    elif ib_link.name == '' and ib_link.ob_comb_link.name != '':
        return ib_link.ob_comb_link.name
    elif ib_link.name != '' and ib_link.ob_comb_link.name == '':
        return ib_link.name
    else:
        return None


def _getLinktypeForTheCombinedLink(ib_link):
    if ib_link.link_type == ib_link.ob_comb_link.link_type:
        return ib_link.link_type
    else:
        return None


def _getSpeedForTheCombinedLink(ib_link):
    if ib_link.free_speed == ib_link.ob_comb_link.free_speed:
        return ib_link.free_speed
    elif ib_link.free_speed == -1 and ib_link.ob_comb_link.free_speed != -1:
        return ib_link.ob_comb_link.free_speed
    elif ib_link.free_speed != -1 and ib_link.ob_comb_link.free_speed == -1:
        return ib_link.free_speed
    else:
        return None


def _checkLinkAttr(ib_link):
    ob_link = ib_link.ob_comb_link
    if ib_link.name != ob_link.name: return False
    if ib_link.link_type != ob_link.link_type: return False
    if ib_link.free_speed != ob_link.free_speed: return False
    if ib_link.allowed_uses != ob_link.allowed_uses: return False
    if (ib_link.lanes is None and ob_link.lanes is not None) or (ib_link.lanes is not None and ob_link.lanes is None): return False
    if ib_link.from_bidirectional_way != ob_link.from_bidirectional_way: return False
    return True


def _newLinkFromLinks(link_id, up_link, down_link):
    link = Link()
    link.osm_way_id = f'{up_link.osm_way_id};{down_link.osm_way_id}'
    link.link_id = link_id
    link.name = up_link.name
    link.link_type_name = up_link.link_type_name
    link.link_type = up_link.link_type
    link.free_speed = up_link.free_speed
    link.allowed_uses = up_link.allowed_uses
    link.from_bidirectional_way = up_link.from_bidirectional_way

    link.from_node = up_link.from_node
    link.to_node = down_link.to_node
    link.from_node.outgoing_link_list.remove(up_link)
    link.to_node.incoming_link_list.remove(down_link)
    link.from_node.outgoing_link_list.append(link)
    link.to_node.incoming_link_list.append(link)

    link.geometry = ops.linemerge([up_link.geometry, down_link.geometry])
    link.length = up_link.length + down_link.length

    up_lanes_list, down_lanes_list = up_link.lanes_list, down_link.lanes_list
    new_down_link_lanes_change_point_list = [point + up_link.length for point in down_link.lanes_change_point_list]
    if up_lanes_list[-1] == down_lanes_list[0]:
        link.lanes_list = up_lanes_list + down_lanes_list[1:]
        link.lanes_change_point_list = up_link.lanes_change_point_list + new_down_link_lanes_change_point_list
    else:
        link.lanes_list = up_lanes_list + down_lanes_list
        link.lanes_change_point_list = up_link.lanes_change_point_list + [up_link.length] + new_down_link_lanes_change_point_list
    link.lanes = link.lanes_list[0]
    return link


def _combLinks(network):

    removal_node_set = set()
    removal_link_set = set()

    for node_id, node in network.node_dict.items():
        if 'traffic_signals' in node.osm_highway: continue

        # check topology
        if len(node.incoming_link_list) == 1 and len(node.outgoing_link_list) == 1:
            topology_flag = _checkTpologyForTwoDegreeNodes(node)
        elif len(node.incoming_link_list) == 2 and len(node.outgoing_link_list) == 2:
            topology_flag = _checkTpologyForFourDegreeNodes(node)
        else:
            topology_flag = False
        if not topology_flag: continue

        # check link attributes, name, link_type, free_speed
        attr_flag = True
        for ib_link in node.incoming_link_list:
            attr_flag = _checkLinkAttr(ib_link)
            if not attr_flag: break
        if not attr_flag: continue

        node.valid = False
        removal_node_set.add(node.node_id)
        for ib_link_idx, ib_link in enumerate(node.incoming_link_list):
            ob_comb_link = ib_link.ob_comb_link
            new_link = _newLinkFromLinks(network.max_link_id, ib_link, ob_comb_link)
            network.link_dict[new_link.link_id] = new_link
            network.max_link_id += 1
            ib_link.valid = False
            ob_comb_link.valid = False
            removal_link_set.add(ib_link.link_id)
            removal_link_set.add(ob_comb_link.link_id)

    for node_id in removal_node_set: del network.node_dict[node_id]
    for link_id in removal_link_set: del network.link_dict[link_id]


def _generateSegments(network):
    segment_list = []
    max_segment_id = network.max_segment_id
    for link_id, link in network.link_dict.items():
        number_of_lanes_change_points = len(link.lanes_change_point_list)
        if number_of_lanes_change_points == 0: continue

        for i in range(number_of_lanes_change_points-1):
            segment = Segment()
            segment.segment_id = max_segment_id
            segment.link = link
            segment.start_lr = link.lanes_change_point_list[i]
            segment.end_lr = link.lanes_change_point_list[i+1]
            segment.l_lanes_added = link.lanes_list[i+1] - link.lanes
            segment_list.append(segment)
            max_segment_id += 1

        segment = Segment()
        segment.segment_id = max_segment_id
        segment.link = link
        segment.start_lr = link.lanes_change_point_list[-1]
        segment.end_lr = link.length
        segment.l_lanes_added = link.lanes_list[-1] - link.lanes
        segment.r_lanes_added = 0
        segment_list.append(segment)
        max_segment_id += 1
    network.segment_list += segment_list
    network.max_segment_id = max_segment_id


def combineShortLinks(network):
    network.link_combined = True
    _combLinks(network)
    _generateSegments(network)



import logging

print_log = False
print_log_level = logging.WARNING

osm_highway_type_dict = {'motorway': 'motorway',
                         'motorway_link': 'motorway',
                         'trunk': 'trunk',
                         'trunk_link': 'trunk',
                         'primary': 'primary',
                         'primary_link': 'primary',
                         'secondary': 'secondary',
                         'secondary_link': 'secondary',
                         'tertiary': 'tertiary',
                         'tertiary_link': 'tertiary',
                         'residential': 'residential',
                         'residential_link': 'residential',
                         'service': 'service',
                         'services': 'service',
                         'cycleway':'cycleway',
                         'footway': 'footway',
                         'pedestrian': 'footway',
                         'steps': 'footway',
                         'track': 'track',
                         'unclassified': 'unclassified'}

link_type_no_dict = {'motorway':1, 'trunk':2, 'primary':3, 'secondary':4, 'tertiary':5, 'residential':6, 'service':7,
                     'cycleway':8, 'footway':9, 'track':10, 'unclassified':11, 'connector':20, 'railway':30, 'aeroway':31}

default_lanes_dict = {'motorway': 4, 'trunk': 3, 'primary': 3, 'secondary': 2, 'tertiary': 2, 'residential': 1, 'service': 1,
                      'cycleway':1, 'footway':1, 'track':1, 'unclassified': 1, 'connector': 2}
default_speed_dict = {'motorway': 59, 'trunk': 39, 'primary': 39, 'secondary': 39, 'tertiary': 29, 'residential': 29, 'service': 29,
                      'cycleway':9, 'footway':4, 'track':29, 'unclassified': 29, 'connector':59}


default_oneway_flag_dict = {'motorway': False,'trunk':False,'primary':False,'secondary':False,'tertiary':False,
                            'residential':False,'service':False,'cycleway':True, 'footway':True,'track': True,
                            'unclassified':False, 'connector':False, 'railway':True, 'aeroway':True}




default_int_buffer = 20.0           # meter
lonlat_precision = 7
xy_precision = 1





from .autoconintd import *
from .autoconm import *
from .classes import Movement


def _getMovementStr(ib_start, ib_end, ob_end):
    angle_ib = math.atan2(ib_end[1] - ib_start[1], ib_end[0] - ib_start[0])
    if -0.75 * math.pi <= angle_ib < -0.25 * math.pi:
        direction = 'SB'
    elif -0.25 * math.pi <= angle_ib < 0.25 * math.pi:
        direction = 'EB'
    elif 0.25 * math.pi <= angle_ib < 0.75 * math.pi:
        direction = 'NB'
    else:
        direction = 'WB'

    angle_ob = math.atan2(ob_end[1] - ib_end[1], ob_end[0] - ib_end[0])
    # angle_ob = math.atan2(ob_end[1] - ob_start[1], ob_end[0] - ob_start[0])
    angle = angle_ob - angle_ib
    if angle < -1 * math.pi:
        angle += 2 * math.pi
    if angle > math.pi:
        angle -= 2 * math.pi

    if -0.25 * math.pi <= angle <= 0.25 * math.pi:
        mvmt = 'T'
    elif angle < -0.25 * math.pi:
        mvmt = 'R'
    elif angle <= 0.75 * math.pi:
        mvmt = 'L'
    else:
        mvmt = 'U'

    mvmt_str = direction + mvmt
    return mvmt_str


def _generateMovementsForOneNode(node):
    number_of_incoming_links = len(node.incoming_link_list)
    number_of_outgoing_links = len(node.outgoing_link_list)

    if number_of_incoming_links == 0 or number_of_outgoing_links == 0: return []
    if number_of_incoming_links == 1 and number_of_outgoing_links == 1: return []
    if number_of_incoming_links == 2 and number_of_outgoing_links == 2: return []

    mvmt_list = []

    if number_of_outgoing_links == 1:
        # merge
        ob_link = node.outgoing_link_list[0]
        ib_link_list = []
        for ib_link in node.incoming_link_list:
            if ib_link.from_node.node_id != ob_link.to_node.node_id:
                ib_link_list.append(ib_link)

        if len(ib_link_list) == 0: return []

        CAutoConnectorM.ob_link, CAutoConnectorM.ib_link_list = ob_link, ib_link_list
        connection_list = CAutoConnectorM.buildConnector()

        for ib_link_no, ib_link in enumerate(ib_link_list):
            ib_lane_index_start = connection_list[ib_link_no][0][0] + 1
            ib_lane_index_end = connection_list[ib_link_no][0][1] + 1
            ob_lane_index_start = connection_list[ib_link_no][1][0] + 1
            ob_lane_index_end = connection_list[ib_link_no][1][1] + 1

            ib_lane = '|'.join(list(map(str,range(ib_lane_index_start, ib_lane_index_end+1))))
            ob_lane = '|'.join(list(map(str, range(ob_lane_index_start, ob_lane_index_end + 1))))
            number_of_lanes = ib_lane_index_end - ib_lane_index_start + 1

            ib_start, ib_end = ib_link.geometry_xy.coords[0], ib_link.geometry_xy.coords[-1]
            ob_start, ob_end = ob_link.geometry_xy.coords[0], ob_link.geometry_xy.coords[-1]
            movement_str = _getMovementStr(ib_start, ib_end, ob_end)

            mvmt = Movement()
            mvmt.node = node
            mvmt.ib_link = ib_link
            mvmt.ib_lane = ib_lane
            mvmt.ob_link = ob_link
            mvmt.ob_lane = ob_lane
            mvmt.lanes = number_of_lanes
            mvmt.movement_str = movement_str
            mvmt_list.append(mvmt)
    else:
        # diverge and intersections
        for ib_link in node.incoming_link_list:
            ob_link_list = []
            for ob_link in node.outgoing_link_list:
                if ib_link.from_node.node_id != ob_link.to_node.node_id:
                    ob_link_list.append(ob_link)

            if len(ob_link_list) == 0: continue

            CAutoConnectorIntD.ib_link, CAutoConnectorIntD.ob_link_list = ib_link, ob_link_list
            connection_list = CAutoConnectorIntD.buildConnector()

            for ob_link_no, ob_link in enumerate(ob_link_list):
                ib_lane_index_start = connection_list[ob_link_no][0][0] + 1
                ib_lane_index_end = connection_list[ob_link_no][0][1] + 1
                ob_lane_index_start = connection_list[ob_link_no][1][0] + 1
                ob_lane_index_end = connection_list[ob_link_no][1][1] + 1

                ib_lane = '|'.join(list(map(str, range(ib_lane_index_start, ib_lane_index_end + 1))))
                ob_lane = '|'.join(list(map(str, range(ob_lane_index_start, ob_lane_index_end + 1))))
                number_of_lanes = ib_lane_index_end - ib_lane_index_start + 1

                ib_start, ib_end = ib_link.geometry_xy.coords[0], ib_link.geometry_xy.coords[-1]
                ob_start, ob_end = ob_link.geometry_xy.coords[0], ob_link.geometry_xy.coords[-1]
                movement_str = _getMovementStr(ib_start, ib_end, ob_end)

                mvmt = Movement()
                mvmt.node = node
                mvmt.ib_link = ib_link
                mvmt.ib_lane = ib_lane
                mvmt.ob_link = ob_link
                mvmt.ob_lane = ob_lane
                mvmt.lanes = number_of_lanes
                mvmt.movement_str = movement_str
                mvmt_list.append(mvmt)

    return mvmt_list


def generateMovements(net):
    if not net.default_lanes:
        print('Movement generation needs complete link lanes information.\n'
              'Argument "default_lanes" must be set as "True" in the function "getNetFromOSMFile()" or'
              '"getNetFromPBFFile()" to give a default vaule to links without lanes information.\n'
              'Function "generateMovements()" will be skipped in this run')
        return


    movement_list_all = []
    number_of_movements = 0

    for node_id, node in net.node_dict.items():
        movement_list = _generateMovementsForOneNode(node)
        for movement in movement_list:
            movement.movement_id = number_of_movements
            movement_list_all.append(movement)
            number_of_movements += 1

    net.movement_list = movement_list_all


from .readfile import *
from .combine_links import *
from .complex_intersection import *
from .wayfilters import *
from .pois import *
from .coordconvertor import from_latlon
import re
from shapely import wkt
import numpy as np


# todo: highway type britain-and-ireland

def _newLinkFromWay(link_id, way, direction, ref_node_list):
    link = Link()
    link.osm_way_id = way.osm_way_id
    link.link_id = link_id
    link.name = way.name
    link.link_type_name = way.link_type_name
    link.link_type = way.link_type
    link.free_speed = way.maxspeed
    link.allowed_uses = way.allowed_uses
    if not way.oneway: link.from_bidirectional_way = True

    if way.oneway:
        link.lanes_list = [way.lanes]
    else:
        if direction == 1:
            if way.forward_lanes is not None:
                link.lanes_list = [way.forward_lanes]
            elif way.lanes is not None:
                link.lanes_list = [math.ceil(way.lanes / 2)]
            else:
                link.lanes_list = [way.lanes]
        else:
            if way.backward_lanes is not None:
                link.lanes_list = [way.backward_lanes]
            elif way.lanes is not None:
                link.lanes_list = [math.ceil(way.lanes / 2)]
            else:
                link.lanes_list = [way.lanes]

    link.lanes = link.lanes_list[0]
    link.from_node = ref_node_list[0]
    link.to_node = ref_node_list[-1]
    link.from_node.outgoing_link_list.append(link)
    link.to_node.incoming_link_list.append(link)
    link.geometry, link.geometry_xy = getLineFromRefNodes(ref_node_list)
    link.calculateLength()
    return link


def _createNodeOnBoundary(node_in,node_outside,network):
    node = Node()
    node.node_id = network.max_node_id
    line = network.bounds.intersection(geometry.LineString([node_in.geometry,node_outside.geometry]))
    lon, lat = line.coords[1]
    node.geometry = geometry.Point((round(lon,lonlat_precision),round(lat,lonlat_precision)))
    x, y = from_latlon(lon, lat, network.central_lon)
    node.geometry_xy = geometry.Point((round(x,xy_precision),round(y,xy_precision)))
    node.is_crossing = True
    node.notes = 'boundary node created by osm2gmns'
    network.node_dict[node.node_id] = node
    network.max_node_id += 1
    return node


def _getSegmentNodeList(way, segment_no, network):
    m_segment_node_list = way.segment_node_list[segment_no]
    if way.is_reversed: m_segment_node_list = list(reversed(m_segment_node_list))
    number_of_nodes = len(m_segment_node_list)

    m_segment_node_list_group = []

    if m_segment_node_list[0].in_region:
        idx_first_outside = -1
        for idx, node in enumerate(m_segment_node_list):
            if not node.in_region:
                idx_first_outside = idx
                break

        if idx_first_outside == -1:
            m_segment_node_list_group.append(m_segment_node_list)
            return m_segment_node_list_group
        else:
            new_node = _createNodeOnBoundary(m_segment_node_list[idx_first_outside-1],m_segment_node_list[idx_first_outside], network)
            m_segment_node_list_group.append(m_segment_node_list[:idx_first_outside] + [new_node])

    if m_segment_node_list[-1].in_region:
        idx_last_outside = -1
        for idx in range(number_of_nodes-2,-1,-1):
            if not m_segment_node_list[idx].in_region:
                idx_last_outside = idx
                break
        new_node = _createNodeOnBoundary(m_segment_node_list[idx_last_outside+1],m_segment_node_list[idx_last_outside], network)
        m_segment_node_list_group.append([new_node] + m_segment_node_list[idx_last_outside+1:])

    return m_segment_node_list_group


def _createLinks(network, link_way_list):
    link_dict = {}
    max_link_id = network.max_link_id
    for way in link_way_list:
        if way.is_pure_cycle: continue
        way.getNodeListForSegments()
        for segment_no in range(way.number_of_segments):
            m_segment_node_list_group = _getSegmentNodeList(way, segment_no, network)
            for m_segment_node_list in m_segment_node_list_group:
                link = _newLinkFromWay(max_link_id, way, 1, m_segment_node_list)
                link_dict[link.link_id] = link
                max_link_id += 1
                if not way.oneway:
                    linkb = _newLinkFromWay(max_link_id, way, -1, list(reversed(m_segment_node_list)))
                    link_dict[linkb.link_id] = linkb
                    max_link_id += 1
    network.link_dict = link_dict
    network.max_link_id = max_link_id


def _identifyCrossingNodes(link_way_list):
    used_node_set = set()
    for way in link_way_list:
        way.ref_node_list[0].is_crossing = True
        way.ref_node_list[-1].is_crossing = True
        for node in way.ref_node_list[1:-1]:
            if node in used_node_set:
                node.is_crossing = True
            else:
                used_node_set.add(node)


def _identifyPureCycleWays(link_way_list):
    for way in link_way_list:
        if way.is_cycle:
            way.is_pure_cycle = True
            for node in way.ref_node_list[1:-1]:
                if node.is_crossing:
                    way.is_pure_cycle = False
                    break


def _getNetworkNodes(network):
    node_dict = {}
    max_node_id = network.max_node_id
    for osm_node_id, node in network.osm_node_dict.items():
        if node.is_crossing and node.in_region:
            node.node_id = max_node_id
            node_dict[node.node_id] = node
            max_node_id += 1
    network.node_dict = node_dict
    network.max_node_id = max_node_id


def _removeIsolated(network,min_nodes):
    node_list = []
    node_to_idx_dict = {}
    for idx, (node_id,node) in enumerate(network.node_dict.items()):
        node_list.append(node)
        node_to_idx_dict[node] = idx

    number_of_nodes = len(node_list)
    node_group_id_list = [-1] * number_of_nodes

    group_id = 0
    start_idx = 0

    while True:
        unprocessed_node_list = [node_list[start_idx]]
        node_group_id_list[start_idx] = group_id
        while unprocessed_node_list:
            node = unprocessed_node_list.pop()
            for ob_link in node.outgoing_link_list:
                ob_node = ob_link.to_node
                if node_group_id_list[node_to_idx_dict[ob_node]] == -1:
                    node_group_id_list[node_to_idx_dict[ob_node]] = group_id
                    unprocessed_node_list.append(ob_node)

            for ib_link in node.incoming_link_list:
                ib_node = ib_link.from_node
                if node_group_id_list[node_to_idx_dict[ib_node]] == -1:
                    node_group_id_list[node_to_idx_dict[ib_node]] = group_id
                    unprocessed_node_list.append(ib_node)

        unreachable_node_exits = False
        for idx in range(start_idx+1,number_of_nodes):
            if node_group_id_list[idx] == -1:
                unreachable_node_exits = True
                break

        if unreachable_node_exits:
            start_idx = idx
            group_id += 1
        else:
            break

    group_id_set = set(node_group_id_list)
    group_isolated_dict = {}
    for group_id in group_id_set:
        group_size = node_group_id_list.count(group_id)
        if group_size < min_nodes:
            group_isolated_dict[group_id] = True
        else:
            group_isolated_dict[group_id] = False

    removal_link_set = set()
    for idx, node in enumerate(node_list):
        if group_isolated_dict[node_group_id_list[idx]]:
            del network.node_dict[node.node_id]
            for ob_link in node.outgoing_link_list: removal_link_set.add(ob_link)
            for ib_link in node.incoming_link_list: removal_link_set.add(ib_link)
    for link in removal_link_set:
        del network.link_dict[link.link_id]


def _getValidNetworkType(network_type):
    if isinstance(network_type,str):
        network_type_temp = (network_type,)
    else:
        network_type_temp = network_type

    network_type_valid = []
    for net_type in network_type_temp:
        if net_type not in network_type_all:
            print(f'network type \'{net_type}\' does not belong to {network_type_all}, it will be skipped')
        else:
            network_type_valid.append(net_type)
    return network_type_valid


def _updateDefaultLaneSpeed(default_lanes, default_speed, network):
    if default_lanes:
        if isinstance(default_lanes,bool):
            network.default_lanes = default_lanes_dict
        elif isinstance(default_lanes,dict):
            network.default_lanes = default_lanes
            for link_type_name, lanes in default_lanes_dict.items():
                if link_type_name not in default_lanes.keys():
                    network.default_lanes[link_type_name] = lanes
        else:
            print('unsupported type for argument default_lanes, has set it to default value: False')
    if default_speed:
        if isinstance(default_speed,bool):
            network.default_speed = default_speed_dict
        elif isinstance(default_speed,dict):
            network.default_speed = default_speed
            for link_type_name, speed in default_speed_dict.items():
                if link_type_name not in default_speed.keys():
                    network.default_speed[link_type_name] = speed
        else:
            print('unsupported type for argument default_lanes, has set it to default value: False')


def _parseNodes(network, nodes, strict_mode):
    osm_node_dict = {}
    osm_node_id_list = []
    osm_node_coord_list = []

    for osm_node in nodes:
        node = Node()
        node.osm_node_id = str(osm_node.id)
        lon,lat = osm_node.lonlat
        node.geometry = geometry.Point((round(lon,lonlat_precision),round(lat,lonlat_precision)))

        if strict_mode:
            if not node.geometry.within(network.bounds):
                node.in_region = False

        osm_node_id_list.append(node.osm_node_id)
        osm_node_coord_list.append((lon,lat))

        tags = osm_node.tags
        if 'highway' in tags.keys():
            node.osm_highway = tags['highway']
        if 'signal' in node.osm_highway: node.ctrl_type = 1         # todo: check signalized tag

        osm_node_dict[node.osm_node_id] = node

    coord_array = np.array(osm_node_coord_list)
    central_lon = coord_array[:,0].mean()
    network.central_lon = central_lon
    xs, ys = from_latlon(coord_array[:,0], coord_array[:,1], central_lon)
    for node_no, node_id in enumerate(osm_node_id_list):
        node = osm_node_dict[node_id]
        node.geometry_xy = geometry.Point((round(xs[node_no],xy_precision),round(ys[node_no],xy_precision)))

    network.osm_node_dict = osm_node_dict


def _parseWays(network, ways, relations, network_type, POIs):
    osm_way_dict = {}
    osm_node_dict = network.osm_node_dict

    for osm_way in ways:
        way = Way()
        way.osm_way_id = str(osm_way.id)

        valid = True
        for ref_node_id in osm_way.refs:
            try:
                way.ref_node_list.append(osm_node_dict[str(ref_node_id)])
            except KeyError:
                valid = False
                print(f'  warning: ref node {ref_node_id} in way {way.osm_way_id} is not defined, way {way.osm_way_id} will not be imported')
        if not valid:
            continue

        tags = osm_way.tags
        if 'highway' in tags.keys():
            way.highway = tags['highway']
        if 'railway' in tags.keys():
            way.railway = tags['railway']
        if 'aeroway' in tags.keys():
            way.aeroway = tags['aeroway']
        if 'lanes' in tags.keys():
            lanes = re.findall(r'\d+\.?\d*', tags['lanes'])
            if len(lanes) > 0:
                way.lanes = int(float(lanes[0]))  # in case of decimals
            else:
                printlog(f'new lanes type detected at way {way.osm_way_id}, {tags["lanes"]}', 'warning')
        if 'lanes:forward' in tags.keys():
            try:
                way.forward_lanes = int(tags['lanes:forward'])
            except:
                pass
        if 'lanes:backward' in tags.keys():
            try:
                way.backward_lanes = int(tags['lanes:backward'])
            except:
                pass
        if 'name' in tags.keys():
            way.name = tags['name']
        if 'maxspeed' in tags.keys():
            speeds = re.findall(r'\d+\.?\d*', tags['maxspeed'])
            if len(speeds) > 0:
                way.maxspeed = speeds[0]
            else:
                printlog(f'new maxspeed type detected at way {way.osm_way_id}, {tags["maxspeed"]}', 'warning')
        if 'oneway' in tags.keys():
            oneway_flag = tags['oneway']
            if oneway_flag == 'yes' or oneway_flag == '1':
                way.oneway = True
            elif oneway_flag == 'no' or oneway_flag == '0':
                way.oneway = False
            elif oneway_flag == '-1':
                way.oneway = True
                way.is_reversed = True
            elif oneway_flag in ['reversible', 'alternating']:
                # todo: reversible, alternating: https://wiki.openstreetmap.org/wiki/Tag:oneway%3Dreversible
                way.oneway = False
            else:
                printlog(f'new lane type detected at way {way.osm_way_id}, {tags["oneway"]}', 'warning')
        if 'area' in tags.keys():
            way.area = tags['area']
        if 'motor_vehicle' in tags.keys():
            way.motor_vehicle = tags['motor_vehicle']
        if 'motorcar' in tags.keys():
            way.motorcar = tags['motorcar']
        if 'service' in tags.keys():
            way.service = tags['service']
        if 'foot' in tags.keys():
            way.foot = tags['foot']
        if 'bicycle' in tags.keys():
            way.bicycle = tags['bicycle']
        if 'building' in tags.keys():
            way.building = tags['building']
        if 'amenity' in tags.keys():
            way.building = tags['amenity']

        osm_way_dict[way.osm_way_id] = way
    network.osm_way_dict = osm_way_dict

    link_way_list = []
    POI_way_list = []
    network_type_set = set(network_type)

    include_railway = True if 'railway' in network_type_set else False
    include_aeroway = True if 'aeroway' in network_type_set else False

    for osm_way_id, way in osm_way_dict.items():
        if way.building or way.amenity:
            POI_way_list.append(way)
        elif way.highway:
            if way.highway in highway_poi_set:
                way.way_poi = way.highway
                POI_way_list.append(way)
                continue
            if way.area and way.area != 'no':
                continue
            if way.highway in negligible_highway_type_set:
                continue

            try:
                way.link_type_name = osm_highway_type_dict[way.highway]
                way.link_type = link_type_no_dict[way.link_type_name]
            except KeyError:
                printlog(f'new highway type at way {way.osm_way_id}, {way.highway}', 'warning')
                continue

            if len(way.ref_node_list) < 2: continue

            allowable_agent_type_list = getAllowableAgentType(way)
            way.allowable_agent_type_list = list(set(allowable_agent_type_list).intersection(network_type_set))
            if len(way.allowable_agent_type_list) == 0:
                continue
            way.allowed_uses = ';'.join(way.allowable_agent_type_list)

            if way.ref_node_list[0] is way.ref_node_list[-1]:
                way.is_cycle = True
            if way.oneway is None:
                way.oneway = default_oneway_flag_dict[way.link_type_name]
            if way.lanes is None:
                if network.default_lanes:
                    way.lanes = network.default_lanes[way.link_type_name]
            if way.maxspeed is None:
                if network.default_speed:
                    way.maxspeed = network.default_speed[way.link_type_name]
            link_way_list.append(way)

        elif way.railway:
            if not include_railway: continue

            if way.railway in railway_poi_set:
                way.way_poi = way.railway
                POI_way_list.append(way)
                continue
            if way.area and way.area != 'no':
                continue
            if way.railway in negligible_railway_type_set:
                continue

            way.link_type_name = way.railway
            way.link_type = link_type_no_dict['railway']
            if way.oneway is None:
                way.oneway = default_oneway_flag_dict['railway']
            link_way_list.append(way)

        elif way.aeroway:
            if not include_aeroway: continue

            if way.aeroway in aeroway_poi_set:
                way.way_poi = way.aeroway
                POI_way_list.append(way)
                continue
            if way.area and way.area != 'no':
                continue
            if way.aeroway in negligible_aeroway_type_set:
                continue

            way.link_type_name = way.aeroway
            way.link_type = link_type_no_dict['aeroway']
            if way.oneway is None:
                way.oneway = default_oneway_flag_dict['aeroway']
            link_way_list.append(way)

        else:
            pass

    _identifyCrossingNodes(link_way_list)
    _getNetworkNodes(network)               # osm node -> node

    _identifyPureCycleWays(link_way_list)
    _createLinks(network, link_way_list)

    if POIs:
        POI_list = generatePOIs(POI_way_list, relations, network)
        network.POI_list = POI_list


def _parseOSM(network, nodes, ways, relations, strict_mode, network_type, POIs):
    _parseNodes(network, nodes, strict_mode)
    _parseWays(network, ways, relations, network_type, POIs)


def _buildNet(netdata,network_type_valid, POIs,strict_mode, min_nodes, combine, int_buffer, bbox, default_lanes, default_speed):
    network = Network()

    _updateDefaultLaneSpeed(default_lanes, default_speed, network)

    bounds, nodes, ways, relations = netdata['bounds'], netdata['nodes'], netdata['ways'], netdata['relations']

    minlat, minlon, maxlat, maxlon = bbox if bbox else bounds['minlat'], bounds['minlon'], bounds['maxlat'], bounds['maxlon']
    network.bounds = geometry.Polygon([(minlon, maxlat), (maxlon, maxlat), (maxlon, minlat), (minlon, minlat)])

    _parseOSM(network, nodes, ways, relations, strict_mode, network_type_valid, POIs)

    # remove isolated nodes and links
    if min_nodes > 1: _removeIsolated(network, min_nodes)

    # combine adjacent links at two-degree nodes
    if combine: combineShortLinks(network)

    # identify complex intersections which contains multiple nodes
    identifyComplexIntersections(network, int_buffer)

    return network


def _getNetClassDict(network_type_valid):
    # was used to reduce memory usage
    net_class_dict = {'highway':False, 'railway':False, 'aeroway':False}
    if 'auto' in network_type_valid or 'bike' in network_type_valid or 'walk' in network_type_valid:
        net_class_dict['highway'] = True
    if 'railway' in network_type_valid:
        net_class_dict['railway'] = True
    if 'aeroway' in network_type_valid:
        net_class_dict['aeroway'] = True
    return net_class_dict


def getNetFromOSMFile(osm_filename='map.osm', network_type=('auto',), POIs=False, strict_mode=True, min_nodes=1,
                      combine=False, int_buffer=default_int_buffer, bbox=None, default_lanes=False, default_speed=False):
    """
    :param osm_filename:
    :param network_type:
    :param POIs:
    :param strict_mode:
    :param min_nodes:
    :param combine:
    :param int_buffer:
    :param bbox: (minlat, minlon, maxlat, maxlon)
    :param default_lanes: True; False; Dict
    :param default_speed: True; False; Dict
    :return: a network instance
    """

    network_type_valid = _getValidNetworkType(network_type)
    netdata = readXMLFile(osm_filename)
    network = _buildNet(netdata,network_type_valid, POIs,strict_mode, min_nodes, combine, int_buffer, bbox, default_lanes, default_speed)
    return network


def getNetFromPBFFile(pbf_filename='map.osm.pbf', network_type=('auto',), POIs=False, strict_mode=True, min_nodes=1,
                      combine=False, int_buffer=default_int_buffer, bbox=None, default_lanes=False, default_speed=False):
    """
    :param pbf_filename:
    :param network_type:
    :param POIs:
    :param strict_mode:
    :param min_nodes:
    :param combine:
    :param int_buffer:
    :param bbox: (minlat, minlon, maxlat, maxlon)
    :param default_lanes: True; False; Dict
    :param default_speed: True; False; Dict
    :return: a network instance
    """

    network_type_valid = _getValidNetworkType(network_type)
    netdata = readPBFFile(pbf_filename)
    network = _buildNet(netdata,network_type_valid, POIs,strict_mode, min_nodes, combine, int_buffer, bbox, default_lanes, default_speed)
    return network


def getNetFromCSV(folder=''):
    node_data, link_data = readCSVFile(folder)
    network = Network()
    max_node_id = 0
    max_link_id = 0

    for i in range(len(node_data)):
        node = Node()
        node.node_id = node_data.loc[i, 'node_id']
        if node.node_id > max_node_id: max_node_id = node.node_id
        osm_node_id = node_data.loc[i, 'osm_node_id']
        if osm_node_id == osm_node_id: node.osm_node_id = osm_node_id
        osm_highway = node_data.loc[i, 'osm_highway']
        if osm_highway == osm_highway: node.osm_highway = osm_highway
        x_coord, y_coord = node_data.loc[i, 'x_coord'], node_data.loc[i, 'y_coord']
        node.geometry = geometry.Point(x_coord,y_coord)
        main_node_id = node_data.loc[i, 'main_node_id']
        if main_node_id == main_node_id: node.main_node_id = int(main_node_id)
        node.ctrl_type = node_data.loc[i, 'ctrl_type']
        poi_id = node_data.loc[i, 'poi_id']
        if poi_id == poi_id: node.poi_id = int(poi_id)
        network.node_dict[node.node_id] = node
    network.max_node_id = max_node_id + 1

    for i in range(len(link_data)):
        link = Link()
        name = link_data.loc[i, 'name']
        if name == name: link.name = name
        link.link_id = link_data.loc[i, 'link_id']
        if link.link_id > max_link_id: max_link_id = link.link_id
        link.osm_way_id = link_data.loc[i, 'osm_way_id']
        link.from_node = network.node_dict[link_data.loc[i, 'from_node_id']]
        link.to_node = network.node_dict[link_data.loc[i, 'to_node_id']]
        link.length = link_data.loc[i, 'length']
        lanes = link_data.loc[i, 'lanes']
        if lanes == lanes: link.lanes = lanes
        free_speed = link_data.loc[i, 'free_speed']
        if free_speed == free_speed: link.free_speed = free_speed
        link.link_type_name = link_data.loc[i, 'link_type_name']
        link.link_type = link_data.loc[i, 'link_type']
        link.geometry = wkt.loads(link_data.loc[i, 'geometry'])
        from_biway = link_data.loc[i, 'from_biway']
        if from_biway == from_biway:
            if int(from_biway) == 1: link.from_bidirectional_way = True

        network.link_dict[link.link_id] = link
        link.from_node.outgoing_link_list.append(link)
        link.to_node.incoming_link_list.append(link)
    network.max_link_id = max_link_id + 1

    return network



def identifyComplexIntersections(network, int_buffer):
    network.complex_intersection_identified = True

    group_list = []
    group_status = []
    for link_id,link in network.link_dict.items():
        if link.length > int_buffer: continue
        if not('traffic_signals' in link.from_node.osm_highway and 'traffic_signals' in link.to_node.osm_highway): continue
        group_list.append({link.from_node, link.to_node})
        group_status.append(1)

    number_of_valid_groups = sum(group_status)
    while True:
        for group_no1,group1 in enumerate(group_list):
            if group_status[group_no1] == 0: continue
            for group_no2,group2 in enumerate(group_list):
                if group_status[group_no2] == 0: continue
                if group_no1 == group_no2: continue
                if len(group1.intersection(group2)) > 0:
                    group1.update(group2)
                    group_status[group_no2] = 0

        new_number_of_valid_groups = sum(group_status)
        if number_of_valid_groups == new_number_of_valid_groups:
            break
        else:
            number_of_valid_groups = new_number_of_valid_groups

    number_of_main_nodes = 0
    for group_no, group in enumerate(group_list):
        if group_status[group_no] == 0: continue
        for node in group: node.main_node_id = number_of_main_nodes
        number_of_main_nodes += 1








def generateNodeActivityInfo(network):
    node_adjacent_link_type_count_dict = {}

    for link_id, link in network.link_dict.items():
        if link.link_type_name == 'connector': continue

        from_node = link.from_node
        if from_node not in node_adjacent_link_type_count_dict.keys():
            node_adjacent_link_type_count_dict[from_node] = {}
        adjacent_link_type_count_dict = node_adjacent_link_type_count_dict[from_node]

        if link.link_type_name in adjacent_link_type_count_dict.keys():
            adjacent_link_type_count_dict[link.link_type_name] += 1
        else:
            adjacent_link_type_count_dict[link.link_type_name] = 1

        to_node = link.to_node
        if to_node not in node_adjacent_link_type_count_dict.keys():
            node_adjacent_link_type_count_dict[to_node] = {}
        adjacent_link_type_count_dict = node_adjacent_link_type_count_dict[to_node]

        if link.link_type_name in adjacent_link_type_count_dict.keys():
            adjacent_link_type_count_dict[link.link_type_name] += 1
        else:
            adjacent_link_type_count_dict[link.link_type_name] = 1

    for node_id, node in network.node_dict.items():
        if node.poi_id is not None:
            node.activity_type = 'poi'
        else:
            adjacent_link_type_count_dict = node_adjacent_link_type_count_dict[node]
            max_count_type = ''
            max_count = 0
            for link_type_name, count in adjacent_link_type_count_dict.items():
                if count > max_count:
                    max_count = count
                    max_count_type = link_type_name
            node.activity_type = max_count_type


    for node_id, node in network.node_dict.items():
        node.is_boundary = False
        if node.activity_type == 'poi': continue
        if (len(node.incoming_link_list) == 0) or (len(node.outgoing_link_list) == 0):
            node.is_boundary = True
            continue

        if (len(node.incoming_link_list) == 1) and (len(node.outgoing_link_list) == 1):
            ib_link = node.incoming_link_list[0]
            ob_link = node.outgoing_link_list[0]
            if ib_link.from_node is ob_link.to_node:
                node.is_boundary = True


# Copyright (C) 2012 Tobias Bieniek <Tobias.Bieniek@gmx.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


try:
    import numpy as mathlib
    use_numpy = True
except ImportError:
    import math as mathlib
    use_numpy = False


K0 = 0.9996

E = 0.00669438
E2 = E * E
E3 = E2 * E
E_P2 = E / (1.0 - E)

SQRT_E = mathlib.sqrt(1 - E)
_E = (1 - SQRT_E) / (1 + SQRT_E)
_E2 = _E * _E
_E3 = _E2 * _E
_E4 = _E3 * _E
_E5 = _E4 * _E

M1 = (1 - E / 4 - 3 * E2 / 64 - 5 * E3 / 256)
M2 = (3 * E / 8 + 3 * E2 / 32 + 45 * E3 / 1024)
M3 = (15 * E2 / 256 + 45 * E3 / 1024)
M4 = (35 * E3 / 3072)

P2 = (3. / 2 * _E - 27. / 32 * _E3 + 269. / 512 * _E5)
P3 = (21. / 16 * _E2 - 55. / 32 * _E4)
P4 = (151. / 96 * _E3 - 417. / 128 * _E5)
P5 = (1097. / 512 * _E4)

R = 6378137


def mixed_signs(x):
    return use_numpy and mathlib.min(x) < 0 <= mathlib.max(x)


def negative(x):
    if use_numpy:
        return mathlib.max(x) < 0
    return x < 0


def from_latlon(longitude, latitude, central_longitude):
    central_lon_rad = mathlib.radians(central_longitude)

    lat_rad = mathlib.radians(latitude)
    lat_sin = mathlib.sin(lat_rad)
    lat_cos = mathlib.cos(lat_rad)

    lat_tan = lat_sin / lat_cos
    lat_tan2 = lat_tan * lat_tan
    lat_tan4 = lat_tan2 * lat_tan2

    lon_rad = mathlib.radians(longitude)

    n = R / mathlib.sqrt(1 - E * lat_sin ** 2)
    c = E_P2 * lat_cos ** 2

    a = lat_cos * (lon_rad - central_lon_rad)
    a2 = a * a
    a3 = a2 * a
    a4 = a3 * a
    a5 = a4 * a
    a6 = a5 * a

    m = R * (M1 * lat_rad - M2 * mathlib.sin(2 * lat_rad) + M3 * mathlib.sin(4 * lat_rad) - M4 * mathlib.sin(6 * lat_rad))
    easting = K0 * n * (a + a3 / 6 * (1 - lat_tan2 + c) + a5 / 120 * (5 - 18 * lat_tan2 + lat_tan4 + 72 * c - 58 * E_P2)) + 500000
    northing = K0 * (m + n * lat_tan * (a2 / 2 + a4 / 24 * (5 - lat_tan2 + 9 * c + 4 * c ** 2) + a6 / 720 * (61 - 58 * lat_tan2 + lat_tan4 + 600 * c - 330 * E_P2)))

    if mixed_signs(latitude):
        raise ValueError("latitudes must all have the same sign")
    elif negative(latitude):
        northing += 10000000

    return easting, northing


import pandas as pd
import os
import locale
from .util import *


def _getBounds(element):
    try:
        minlat = float(element.attrib['minlat'])
    except KeyError:
        minlat = -90.0
    try:
        minlon = float(element.attrib['minlon'])
    except KeyError:
        minlon = -180.0
    try:
        maxlat = float(element.attrib['maxlat'])
    except KeyError:
        maxlat = 90.0
    try:
        maxlon = float(element.attrib['maxlon'])
    except KeyError:
        maxlon = 180.0
    return {'minlat':minlat, 'minlon':minlon, 'maxlat':maxlat, 'maxlon':maxlon}


def readXMLFile(osm_filename):
    import xml.etree.cElementTree as ET
    from .file import Node, Way, Relation

    bounds = default_bounds.copy()
    nodes, ways, relations = [], [], []

    osmtree = ET.ElementTree(file=osm_filename)
    osmnet = osmtree.getroot()
    for element in osmnet:
        if element.tag == 'bounds':
            bounds = _getBounds(element)

        elif element.tag == 'node':
            node_id = element.attrib['id']
            lonlat = (float(element.attrib['lon']),float(element.attrib['lat']))
            tags = {}
            for info in element:
                if info.tag == 'tag':
                    tags[info.attrib['k']] = info.attrib['v']

            node = Node(node_id, tags, lonlat)
            nodes.append(node)

        elif element.tag == 'way':
            way_id = element.attrib['id']
            refs = []
            tags = {}
            for info in element:
                if info.tag == 'nd':
                    ref_node_id = info.attrib['ref']
                    refs.append(ref_node_id)
                elif info.tag == 'tag':
                    tags[info.attrib['k']] = info.attrib['v']

            way = Way(way_id,tags, refs)
            ways.append(way)

        elif element.tag == 'relation':
            relation_id = element.attrib['id']
            members = []
            tags = {}
            for info in element:
                if info.tag == 'member':
                    member_type = info.attrib['type']
                    ref_id = info.attrib['ref']
                    member_role = info.attrib['role']
                    members.append((ref_id,member_type,member_role))
                elif info.tag == 'tag':
                    tags[info.attrib['k']] = info.attrib['v']

            relation = Relation(relation_id, tags, members)
            relations.append(relation)

    return {'bounds':bounds, 'nodes':nodes, 'ways':ways, 'relations':relations}


def readPBFFile(pbf_filename):
    from .file import File, Node, Way, Relation

    bounds = default_bounds.copy()
    nodes, ways, relations = [], [], []

    pbf_data = File(pbf_filename)
    for item in pbf_data:
        if isinstance(item,Node):
            nodes.append(item)
        elif isinstance(item, Way):
            ways.append(item)
        elif isinstance(item, Relation):
            relations.append(item)
        else:
            print('unsupported type')

    return {'bounds':bounds, 'nodes':nodes, 'ways':ways, 'relations':relations}


def readCSVFile(folder):
    if folder:
        node_filepath = os.path.join(folder,'node.csv')
        link_filepath = os.path.join(folder,'link.csv')
    else:
        node_filepath = 'node.csv'
        link_filepath = 'link.csv'

    local_encoding = locale.getdefaultlocale()
    try:
        node_data = pd.read_csv(node_filepath,dtype={'osm_node_id':str})
    except UnicodeDecodeError:
        node_data = pd.read_csv(node_filepath,dtype={'osm_node_id':str},encoding=local_encoding[1])
    try:
        link_data = pd.read_csv(link_filepath,dtype={'osm_way_id':str})
    except UnicodeDecodeError:
        link_data = pd.read_csv(link_filepath,dtype={'osm_way_id':str},encoding=local_encoding[1])

    return node_data, link_data



from .network import *
from .complex_intersection import *
from .consolidate_intersections import *
from .activity import *
from .writefile import *
from .movement import generateMovements
from .settings import *



print('osm2gmns, version 0.3.2')

if print_log:
    logging.basicConfig(level=print_log_level,
                        format='%(asctime)s %(filename)s[line:%(lineno)d] %(levelname)s %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S',
                        filename='osm2gmns.log',
                        filemode='w')


# Note: we adopt the filter logic from osmnx (https://github.com/gboeing/osmnx)
# exclude links with tag attributes in the filters

# todo: walk, bike

negligible_highway_type_set = {'path','construction','proposed','raceway','bridleway','rest_area','su','living_street',
                               'road','abandoned','planned','trailhead','stairs','dismantled','disused','razed','access',
                               'corridor','stop'}
negligible_railway_type_set = {'construction','abandoned','disused','proposed','planned','dismantled','razed','ventilation_shaft'}
negligible_aeroway_type_set = set()

highway_poi_set = {'bus_stop','platform'}
railway_poi_set = {'depot','station','workshop','halt','interlocking','junction','spur_junction','terminal','platform'}
aeroway_poi_set = set()

network_type_all = {'auto','bike','walk','railway','aeroway'}


_filter_in = {}
_filters_ex = {}

_filter_in['auto'] = {'motor_vehicle':{'yes'},
                      'motorcar':{'yes'}}

_filter_in['bike'] = {'bicycle':{'yes'}}

_filter_in['walk'] = {'foot':{'yes'}}

_filters_ex['auto'] = {'highway':{'cycleway','footway','pedestrian','steps','track','corridor','elevator','escalator','service'},
                       'motor_vehicle':{'no'},
                       'motorcar':{'no'},
                       'access':{'private'},
                       'service':{'parking','parking_aisle','driveway','private','emergency_access'}}

_filters_ex['bike'] = {'highway':{'footway','steps','corridor','elevator','escalator','motor'},
                       'bicycle':{'no'},
                       'service':{'private'},
                       'access':{'private'}}

_filters_ex['walk'] = {'highway':{'cycleway','motor'},
                       'foot':{'no'},
                       'service':{'private'},
                       'access':{'private'}}

_agent_type_all = ['auto', 'bike', 'walk']



def _checkIn(way, agent_type):
    m_filter_in = _filter_in[agent_type]
    for tag, include_list in m_filter_in.items():
        if eval(f'way.{tag} in include_list'):
            return True
    return None


def _checkEx(way, agent_type):
    m_filter_ex = _filters_ex[agent_type]
    for tag, exclude_list in m_filter_ex.items():
        if eval(f'way.{tag} in exclude_list'):
            return False
    return True


def getAllowableAgentType(way):
    allowable_agent_type_list = []

    for agent_type in _agent_type_all:
        allowed = _checkIn(way, agent_type)
        if allowed:
            allowable_agent_type_list.append(agent_type)
            continue

        allowed = _checkEx(way, agent_type)
        if allowed:
            allowable_agent_type_list.append(agent_type)

    return allowable_agent_type_list



'''
this script is adopted from esy-osm-pbf (https://pypi.org/project/esy-osm-pbf/)

esy.osm.pbf.file
================

Provides classes and functions to work with the OpenStreetMap Protocol Buffers
format. It decodes string tag maps into Python dictionaries as well as scaling
longitude and latitude coordinates.
'''

import sys, zlib, struct, itertools

from collections import namedtuple

from . import osmformat_pb2, fileformat_pb2


if sys.version_info.major >= 3:
    from itertools import accumulate
elif sys.version_info.major == 2 and sys.version_info.minor >= 7:
    # itertools.accumulate is not available in python 2.7
    import operator

    def accumulate(iterable, func=operator.add):
        'Return running totals'
        it = iter(iterable)
        total = next(it)
        yield total
        for element in it:
            total = func(total, element)
            yield total
else:
    raise RuntimeError('Unsupported python version')



def decode_strmap(primitive_block):
    '''Decodes the bytes in the stringtable of `primitive_block` to UTF-8.'''
    return tuple(s.decode('utf8') for s in primitive_block.stringtable.s)


def iter_blocks(file):
    '''Iterates tuples ``(offset, header)`` for each block in the `file`.'''
    ofs = 0
    while True:
        file.seek(ofs)
        data = file.read(4)
        if len(data) < 4:
            return
        header_size, = struct.unpack('>I', data)
        header = fileformat_pb2.BlobHeader()
        header.ParseFromString(file.read(header_size))
        ofs += 4 + header_size
        yield ofs, header
        ofs += header.datasize


def read_blob(file, ofs, header):
    '''Read and decompress a openstreetmap blob from `file`.'''
    file.seek(ofs)
    blob = fileformat_pb2.Blob()
    blob.ParseFromString(file.read(header.datasize))
    if blob.raw:
        return blob.raw
    elif blob.zlib_data:
        return zlib.decompress(blob.zlib_data)
    else:
        raise ValueError('lzma blob not supported')


def parse_tags(strmap, keys_vals):
    '''
    Parse tags from the key/value tuples `keys_vals` and strings from
    `strmap` into a dictionary.
    '''
    tags = {}
    is_value = False
    for idx in keys_vals:
        if idx == 0:
            yield tags
            tags = {}
        elif is_value:
            tags[key] = strmap[idx]
            is_value = False
        else:
            key = strmap[idx]
            is_value = True


def iter_nodes(block, strmap, group):
    dense = group.dense
    if not dense:
        raise ValueError('Only dense nodes are supported')
    granularity = block.granularity or 100
    lat_offset = block.lat_offset or 0
    lon_offset = block.lon_offset or 0
    coord_scale = 0.000000001
    id = lat = lon = tag_pos = 0
    for did, dlat, dlon, tags in zip(
            dense.id, dense.lat, dense.lon,
            parse_tags(strmap, dense.keys_vals)):
        id += did
        lat += coord_scale * (lat_offset + granularity * dlat)
        lon += coord_scale * (lon_offset + granularity * dlon)
        yield (id, tags, (lon, lat))


def iter_ways(block, strmap, group):
    for way in group.ways:
        tags = {
            strmap[k]: strmap[v]
            for k, v in zip(way.keys, way.vals)
        }
        refs = tuple(accumulate(way.refs))
        yield way.id, refs, tags


def iter_relations(block, strmap, group):
    namemap = {}
    for relation in group.relations:
        tags = {
            strmap[k]: strmap[v]
            for k, v in zip(relation.keys, relation.vals)
        }
        refs = tuple(accumulate(relation.memids))
        members = [
            (
                ref,
                namemap.setdefault(
                    rel_type, osmformat_pb2.Relation.MemberType.Name(rel_type)),
                strmap[sid]
            )
            for ref, rel_type, sid in zip(
                    refs, relation.types, relation.roles_sid)
        ]

        yield relation.id, members, tags


Node = namedtuple('Node', ('id', 'tags', 'lonlat'))
Node.__doc__ = '''
A OpenStreetMap `node <https://wiki.openstreetmap.org/wiki/Node>`_.
'''
Node.id.__doc__ = 'Identifier of the node.'
Node.tags.__doc__ = 'A dictonary mapping from tag names to values.'
Node.lonlat.__doc__ = 'A tuple ``(longitude, latitude)`` of coordinates.'

Way = namedtuple('Way', ('id', 'tags', 'refs'))
Way.__doc__ = '''
A OpenStreetMap `way <https://wiki.openstreetmap.org/wiki/Way>`_.
'''
Way.id.__doc__ = 'Identifier of the way.'
Way.tags.__doc__ = 'A dictonary mapping from tag names to values.'
Way.refs.__doc__ = 'A tuple of node identifiers representing this way.'

Relation = namedtuple('Relation', ('id', 'tags', 'members'))
Relation.__doc__ = '''
A OpenStreetMap `relation <https://wiki.openstreetmap.org/wiki/Relation>`_.
'''
Relation.id.__doc__ = 'Identifier of the relation.'
Relation.tags.__doc__ = 'A dictonary mapping from tag names to values.'
Relation.members.__doc__ = '''
A tuple with ``(ref, member_type, role)`` entries describing the relation.
``ref`` is an OSM entry identifier, ``member_type`` is the type of the
referenced entry (either the :class:`Node`, :class:`Way` or :class:`Relation`
type) and ``role`` is a string describing the role.
'''


def iter_primitive_block(primitive_block):
    '''
    Iterates over the entries of a OpenStreetMap primitive block.
    '''
    strmap = decode_strmap(primitive_block)
    for group in primitive_block.primitivegroup:
        for id, tags, lonlat in iter_nodes(primitive_block, strmap, group):
            yield Node(id, tags, lonlat)

        for id, refs, tags in iter_ways(primitive_block, strmap, group):
            yield Way(id, tags, refs)

        for id, members, tags in iter_relations(primitive_block, strmap, group):
            yield Relation(id, tags, members)


class Block(object):
    def __init__(self, file, ofs, header):
        self.file, self.ofs, self.header = file, ofs, header

    def __iter__(self):
        primitive_block = osmformat_pb2.PrimitiveBlock()
        primitive_block.ParseFromString(
            read_blob(self.file.file, self.ofs, self.header))
        return iter_primitive_block(primitive_block)


class File(object):
    '''
    Provides access to OpenStreetMap ``.pbf`` files.

    Iterates over entries (or :attr:`blocks`) of a ``.pbf`` file.
    '''

    def __init__(self, file):
        if isinstance(file, str):
            self.file = open(file, 'rb')
        elif isinstance(file, Path):
            self.file = file.open('rb')
        elif isinstance(file, io.IOBase):
            self.file = file
        else:
            raise ValueError('Unsupported file object {}'.format(file))

    def __enter__(self):
        return self

    def __exit__(self, exc_type=None, exc_value=None, exc_trace=None):
        if hasattr(self.file, 'close'):
            self.file.close()

    def __iter__(self):
        '''
        Iterates over all entries of a ``.pbf`` file.
        '''
        for block in self.blocks:
            yield from block

    @property
    def blocks(self):
        '''
        Iterates over all blocks in a ``.pbf`` file.
        '''
        for ofs, header in iter_blocks(self.file):
            yield Block(self, ofs, header)



import math

class CAutoConnectorIntD:
    default_right_most_lanes = 1        # do not change
    default_left_most_lanes = 1         # do not change

    ib_link = None
    ob_link_list = []
    ob_link_list_sorted = []
    connection_list = []


    @classmethod
    def getSequence(cls):       # order outbound links from left to right
        angle_list = []

        angle_ib = math.atan2(cls.ib_link.geometry_xy.coords[-1][1] - cls.ib_link.geometry_xy.coords[0][1],
                              cls.ib_link.geometry_xy.coords[-1][0] - cls.ib_link.geometry_xy.coords[0][0])
        # angle_ib = math.atan2(cls.ib_link.geometry_list[-1][1] - cls.ib_link.geometry_list[0][1], cls.ib_link.geometry_list[-1][0] - cls.ib_link.geometry_list[0][0])
        for ob_link in cls.ob_link_list:
            angle_ob = math.atan2(ob_link.geometry_xy.coords[-1][1] - ob_link.geometry_xy.coords[0][1],
                                  ob_link.geometry_xy.coords[-1][0] - ob_link.geometry_xy.coords[0][0])
            # angle_ob = math.atan2(ob_link.geometry_list[-1][1] - ob_link.geometry_list[0][1], ob_link.geometry_list[-1][0] - ob_link.geometry_list[0][0])
            angle = angle_ob-angle_ib
            if angle < -1 * math.pi:
                angle += 2 * math.pi
            if angle > math.pi:
                angle -= 2 * math.pi
            angle_list.append(angle)

        cls.ob_link_list_sorted = sorted(cls.ob_link_list, key=lambda x:angle_list[cls.ob_link_list.index(x)],reverse=True)


    @classmethod
    def getLaneConnection(cls):
        cls.connection_list = [[] for _ in range(len(cls.ob_link_list_sorted))]
        ib_lanes = cls.ib_link.lanes

        if ib_lanes == 1:
            left_ob_link = cls.ob_link_list_sorted[0]
            cls.connection_list[cls.ob_link_list.index(left_ob_link)] = [(0, 0),(0, 0)]
            for ob_link in cls.ob_link_list_sorted[1:]:
                cls.connection_list[cls.ob_link_list.index(ob_link)] = [(0, 0), (ob_link.lanes-1, ob_link.lanes-1)]
            return

        if len(cls.ob_link_list_sorted) == 1:
            # only one ob_link, full connection
            ob_link = cls.ob_link_list_sorted[0]
            connection_lanes = min(ib_lanes, ob_link.lanes)
            cls.connection_list[cls.ob_link_list.index(ob_link)] = [(0, connection_lanes - 1), (0, connection_lanes - 1)]

        elif len(cls.ob_link_list_sorted) == 2:
            # two ob_link, default right, remaining left
            left_ob_link = cls.ob_link_list_sorted[0]
            lanes_ib_link_left = ib_lanes - cls.default_left_most_lanes
            lanes_ob_has = left_ob_link.lanes
            connection_lanes = min(lanes_ib_link_left,lanes_ob_has)
            cls.connection_list[cls.ob_link_list.index(left_ob_link)] = [(0, connection_lanes - 1), (0, connection_lanes - 1)]

            right_ob_link = cls.ob_link_list_sorted[-1]
            cls.connection_list[cls.ob_link_list.index(right_ob_link)] = [
                (ib_lanes - cls.default_right_most_lanes, ib_lanes - 1),
                (right_ob_link.lanes - cls.default_right_most_lanes, right_ob_link.lanes - 1)]

        else:
            # >= 3, default left, default right, remaining mid
            left_ob_link = cls.ob_link_list_sorted[0]
            cls.connection_list[cls.ob_link_list.index(left_ob_link)] = [(0,cls.default_left_most_lanes-1), (0,cls.default_left_most_lanes-1)]

            mid_ob_link_list = cls.ob_link_list_sorted[1:-1]
            lanes_assgined_to_each_mid_ob_link = [0] * len(mid_ob_link_list)
            lanes_each_mid_ob_link_left = [link.lanes for link in mid_ob_link_list]

            if ib_lanes - cls.default_left_most_lanes - cls.default_right_most_lanes >= len(mid_ob_link_list):
                lanes_ib_link_left = ib_lanes - cls.default_left_most_lanes - cls.default_right_most_lanes
                start_lane_no = cls.default_left_most_lanes
                while (lanes_ib_link_left > 0) and (sum(lanes_each_mid_ob_link_left) > 0):
                    for ob_link_no, ob_link in enumerate(mid_ob_link_list):
                        if lanes_each_mid_ob_link_left[ob_link_no] == 0: continue
                        if lanes_ib_link_left == 0: break
                        lanes_each_mid_ob_link_left[ob_link_no] -= 1
                        lanes_assgined_to_each_mid_ob_link[ob_link_no] += 1
                        lanes_ib_link_left -= 1
                for ob_link_no, ob_link in enumerate(mid_ob_link_list):
                    cls.connection_list[cls.ob_link_list.index(ob_link)] = [
                        (start_lane_no, start_lane_no + lanes_assgined_to_each_mid_ob_link[ob_link_no] - 1),
                        (ob_link.lanes - lanes_assgined_to_each_mid_ob_link[ob_link_no], ob_link.lanes - 1)]
                    start_lane_no += lanes_assgined_to_each_mid_ob_link[ob_link_no]
            elif ib_lanes < len(mid_ob_link_list):
                lane_no, link_no = -1, -1
                for lane_no in range(ib_lanes-1):
                    link_no = lane_no
                    ob_link = mid_ob_link_list[link_no]
                    cls.connection_list[cls.ob_link_list.index(ob_link)] = [(lane_no, lane_no), (ob_link.lanes - 1, ob_link.lanes - 1)]
                lane_no += 1
                start_link_no = link_no + 1
                for link_no in range(start_link_no, len(mid_ob_link_list)):
                    ob_link = mid_ob_link_list[link_no]
                    cls.connection_list[cls.ob_link_list.index(ob_link)] = [(lane_no, lane_no), (ob_link.lanes - 1, ob_link.lanes - 1)]
            else:
                if ib_lanes - cls.default_left_most_lanes == len(mid_ob_link_list):
                    start_lane_no = cls.default_left_most_lanes
                else:
                    # ib_lanes == len(mid_ob_link_list)
                    start_lane_no = 0
                for ob_link in mid_ob_link_list:
                    cls.connection_list[cls.ob_link_list.index(ob_link)] = [(start_lane_no, start_lane_no), (ob_link.lanes - 1, ob_link.lanes - 1)]
                    start_lane_no += 1

            right_ob_link = cls.ob_link_list_sorted[-1]
            cls.connection_list[cls.ob_link_list.index(right_ob_link)] = [
                (ib_lanes - cls.default_right_most_lanes, ib_lanes - 1),
                (right_ob_link.lanes - cls.default_right_most_lanes, right_ob_link.lanes - 1)]

    @classmethod
    def buildConnector(cls):
        cls.getSequence()
        cls.getLaneConnection()
        return cls.connection_list



import math

class CAutoConnectorM:
    default_right_most_lanes = 1        # do not change
    default_left_most_lanes = 1         # do not change

    ib_link_list = []
    ob_link = None
    ib_link_list_sorted = []
    connection_list = []


    @classmethod
    def getSequence(cls):       # order inbound links from left to right
        angle_list = []
        angle_ob = math.atan2(cls.ob_link.geometry_xy.coords[-1][1] - cls.ob_link.geometry_xy.coords[0][1],
                              cls.ob_link.geometry_xy.coords[-1][0] - cls.ob_link.geometry_xy.coords[0][0])
        # angle_ob = math.atan2(cls.ob_link.geometry_list[-1][1] - cls.ob_link.geometry_list[0][1], cls.ob_link.geometry_list[-1][0] - cls.ob_link.geometry_list[0][0])
        for ib_link in cls.ib_link_list:
            angle_ib = math.atan2(ib_link.geometry_xy.coords[-1][1] - ib_link.geometry_xy.coords[0][1],
                                  ib_link.geometry_xy.coords[-1][0] - ib_link.geometry_xy.coords[0][0])
            # angle_ib = math.atan2(ib_link.geometry_list[-1][1] - ib_link.geometry_list[0][1], ib_link.geometry_list[-1][0] - ib_link.geometry_list[0][0])
            angle = angle_ob-angle_ib
            if angle < -1 * math.pi:
                angle += 2 * math.pi
            if angle > math.pi:
                angle -= 2 * math.pi
            angle_list.append(angle)

        cls.ib_link_list_sorted = sorted(cls.ib_link_list, key=lambda x:angle_list[cls.ib_link_list.index(x)],reverse=True)


    @classmethod
    def getLaneConnection(cls):
        cls.connection_list = [[] for _ in range(len(cls.ib_link_list_sorted))]
        ob_lanes = cls.ob_link.lanes


        left_ib_link = cls.ib_link_list_sorted[0]
        connection_lanes = min(ob_lanes, left_ib_link.lanes)
        cls.connection_list[cls.ib_link_list.index(left_ib_link)] = [(left_ib_link.lanes-connection_lanes, left_ib_link.lanes-1),
                                                                     (0, connection_lanes-1)]        # in, out

        for ib_link in cls.ib_link_list_sorted[1:]:
            connection_lanes = min(ob_lanes, ib_link.lanes)
            cls.connection_list[cls.ib_link_list.index(ib_link)] = [(0, connection_lanes - 1), (ob_lanes-connection_lanes, ob_lanes-1)]  # in, out


    @classmethod
    def buildConnector(cls):
        cls.getSequence()
        cls.getLaneConnection()
        return cls.connection_list



from .util import *


class Segment:
    def __init__(self):
        self.segment_id = 0
        self.link = None
        self.start_lr = 0.0
        self.end_lr = 0.0
        self.l_lanes_added = 0
        self.r_lanes_added = 0


class Movement:
    def __init__(self):
        self.movement_id = 0
        self.node = None
        self.ib_link = None
        self.ib_lane = ''
        self.ob_link = None
        self.ob_lane = ''
        self.lanes = 0
        self.type = ''
        self.penalty = 0.0
        self.capacity = 0.0
        self.ctrl_type = ''
        self.main_node_id = None
        self.movement_str = ''



class Node:
    def __init__(self):
        self.name = None
        self.node_id = 0
        self.osm_node_id = None     # str
        self.geometry = None
        self.geometry_xy = None
        self.main_node_id = None
        self.osm_highway = ''
        self.node_type = ''
        self.ctrl_type = 0
        self.in_region = True
        self.is_crossing = False
        self.is_isolated = False
        self.valid = True
        self.activity_type = ''
        self.is_boundary = None
        self.poi_id = None
        self.notes = ''
        self.incoming_link_list = []
        self.outgoing_link_list = []



class Link:
    def __init__(self):
        self.link_id = 0
        self.osm_way_id = None      # str
        self.name = ''
        self.link_type_name = ''
        self.link_type = 0
        self.from_node = None
        self.to_node = None
        self.lanes = None
        self.lanes_list = []
        self.free_speed = None
        self.length = 0.0
        self.allowed_uses = ''
        self.from_bidirectional_way = False
        self.geometry = None
        self.geometry_xy = None
        self.is_isolated = False
        self.valid = True
        self.ob_comb_link = None
        self.lanes_change_point_list = []

    def calculateLength(self):
        coord_list = self.geometry.coords[:]
        for i in range(len(coord_list)-1):
            from_coord = coord_list[i]
            to_coord = coord_list[i+1]
            self.length += getDistanceFromCoord(from_coord[0], from_coord[1], to_coord[0], to_coord[1]) * 1000


class Way:
    def __init__(self):
        self.osm_way_id = None          # string
        self.highway = ''
        self.railway = ''
        self.aeroway = ''
        self.link_type_name = ''
        self.link_type = 0
        self.name = ''
        self.lanes = None
        self.forward_lanes = None
        self.backward_lanes = None
        self.maxspeed = None
        self.oneway = None

        self.area = None
        self.motor_vehicle = None
        self.motorcar = None
        self.service = None
        self.access = None
        self.foot = None
        self.bicycle = None
        self.building = None
        self.amenity = None
        self.way_poi = None     # used for poi
        self.allowable_agent_type_list = []
        self.allowed_uses = ''

        self.is_reversed = False
        self.is_cycle = False
        self.is_pure_cycle = False          # cycle without crossing nodes
        self.ref_node_list = []
        self.number_of_segments = 0
        self.segment_node_list = []         # ref node sequence for each segment

    def getNodeListForSegments(self):
        number_of_ref_nodes = len(self.ref_node_list)
        last_idx = 0
        idx = 0

        while True:
            m_segment_node_list = [self.ref_node_list[last_idx]]
            for idx in range(last_idx+1,number_of_ref_nodes):
                ref_node = self.ref_node_list[idx]
                m_segment_node_list.append(ref_node)
                if ref_node.is_crossing:
                    last_idx = idx
                    break
            self.segment_node_list.append(m_segment_node_list)
            self.number_of_segments += 1
            if idx == number_of_ref_nodes-1: break


class Relation:
    def __init__(self):
        self.osm_relation_id = None
        self.member_list = []
        self.member_role_list = []
        self.name = ''
        self.building = None
        self.amenity = None


class POI:
    def __init__(self):
        self.poi_id = 0
        self.osm_way_id = None      # str
        self.osm_relation_id = None
        self.name = ''
        self.geometry = None
        self.geometry_xy = None
        self.centroid = None
        self.centroid_xy = None
        self.nearest_node = None
        self.building = None
        self.amenity = None
        self.way = None         # highway,railway,aeroway poi


class Network:
    def __init__(self):
        self.bounds = None
        self.central_lon = 0.0

        self.default_lanes = False
        self.default_speed = False

        self.link_combined = False
        self.complex_intersection_identified = False
        self.consolidated = False
        self.new_id = True

        self.node_coordstr_to_node_dict = {}

        self.osm_node_dict = {}
        self.osm_way_dict = {}

        self.max_node_id = 0
        self.max_link_id = 0
        self.max_segment_id = 0
        self.max_main_node_id = 0
        self.max_poi_id = 0

        self.node_dict = {}
        self.link_dict = {}
        self.segment_list = []
        self.complex_intersection_list = []
        self.POI_list = []
        self.movement_list = []



from .classes import *
from .settings import lonlat_precision, xy_precision


def consolidateComplexIntersections(network):
    node_group_dict = {}
    for node_id, node in network.node_dict.items():
        if node.main_node_id is not None:
            if node.main_node_id in node_group_dict.keys():
                node_group_dict[node.main_node_id].append(node)
            else:
                node_group_dict[node.main_node_id] = [node]

    removal_node_set = set()
    removal_link_set = set()

    for main_node_id, node_group in node_group_dict.items():
        new_node = Node()
        new_node.node_id = network.max_node_id
        new_node.main_node_id = main_node_id
        new_node.ctrl_type = 1
        osm_node_id_list = []
        x_coord_sum, y_coord_sum = 0.0, 0.0
        x_coord_xy_sum, y_coord_xy_sum = 0.0, 0.0

        for node in node_group:
            node.valid = False
            removal_node_set.add(node)

            osm_node_id_list.append(node.osm_node_id)
            x_coord_sum += node.geometry.x
            y_coord_sum += node.geometry.y
            x_coord_xy_sum += node.geometry_xy.x
            y_coord_xy_sum += node.geometry_xy.y

            for link in node.incoming_link_list:
                if link.from_node in node_group:
                    link.valid = False
                    removal_link_set.add(link)
                else:
                    link.to_node = new_node
                    new_node.incoming_link_list.append(link)
            for link in node.outgoing_link_list:
                if link.to_node in node_group:
                    link.valid = False
                    removal_link_set.add(link)
                else:
                    link.from_node = new_node
                    new_node.outgoing_link_list.append(link)

            new_node.osm_highway = node.osm_highway

        new_node.osm_node_id = ';'.join(osm_node_id_list)
        x_coord_ave = round(x_coord_sum / len(node_group),lonlat_precision)
        y_coord_ave = round(y_coord_sum / len(node_group),lonlat_precision)
        new_node.geometry = geometry.Point(x_coord_ave, y_coord_ave)
        x_coord_xy_ave = round(x_coord_xy_sum / len(node_group),xy_precision)
        y_coord_xy_ave = round(y_coord_xy_sum / len(node_group),xy_precision)
        new_node.geometry_xy = geometry.Point(x_coord_xy_ave, y_coord_xy_ave)

        network.node_dict[new_node.node_id] = new_node
        network.max_node_id += 1


    for node in removal_node_set: del network.node_dict[node.node_id]
    for link in removal_link_set: del network.link_dict[link.link_id]


import csv
import os
from shapely import geometry


def outputNetToCSV(network, output_folder='', projection=False):
    if output_folder:
        if not os.path.exists(output_folder): os.mkdir(output_folder)
        node_filepath = os.path.join(output_folder,'node.csv')
        link_filepath = os.path.join(output_folder,'link.csv')
        segment_filepath = os.path.join(output_folder,'segment.csv')
        movement_filepath = os.path.join(output_folder, 'movement.csv')
        structure_filpath = os.path.join(output_folder,'poi.csv')
    else:
        node_filepath = 'node.csv'
        link_filepath = 'link.csv'
        segment_filepath = 'segment.csv'
        movement_filepath = 'movement.csv'
        structure_filpath = 'poi.csv'

    # we use errors='ignore' to make our program compatible with some characters that cannot be encoded by the local encoding,
    # these characters will be discarded

    while True:
        try:
            outfile = open(node_filepath, 'w', newline='',errors='ignore')
            break
        except PermissionError:
            print('node.csv may be locked by other programs. please release it then press Enter to try again')
            input()
    writer = csv.writer(outfile)
    writer.writerow(['name', 'node_id', 'osm_node_id', 'osm_highway', 'zone_id', 'ctrl_type', 'node_type', 'activity_type',
                     'is_boundary', 'x_coord', 'y_coord', 'main_node_id','poi_id','notes'])
    for node_id,node in network.node_dict.items():
        if node.is_boundary is None:
            is_boundary = ''
        else:
            is_boundary = 1 if node.is_boundary else 0
        x_coord = node.geometry_xy.x if projection else node.geometry.x
        y_coord = node.geometry_xy.y if projection else node.geometry.y
        line = [node.name, node.node_id, node.osm_node_id, node.osm_highway, '', node.ctrl_type, '', node.activity_type,
                is_boundary, x_coord, y_coord,node.main_node_id,node.poi_id,node.notes]
        writer.writerow(line)
    outfile.close()

    while True:
        try:
            outfile = open(link_filepath, 'w', newline='',errors='ignore')
            break
        except PermissionError:
            print('link.csv may be locked by other programs. please release it then press Enter to try again')
            input()
    writer = csv.writer(outfile)
    writer.writerow(['name', 'link_id', 'osm_way_id', 'from_node_id', 'to_node_id', 'dir_flag', 'length', 'lanes',
                     'free_speed', 'capacity', 'link_type_name', 'link_type', 'geometry','allowed_uses','from_biway'])
    for link_id, link in network.link_dict.items():
        from_biway = 1 if link.from_bidirectional_way else 0
        geometry_ = link.geometry_xy if projection else link.geometry
        line = [link.name, link.link_id, link.osm_way_id, link.from_node.node_id, link.to_node.node_id, 1, link.length,
                link.lanes, link.free_speed, '', link.link_type_name, link.link_type, geometry_, link.allowed_uses,from_biway]
        writer.writerow(line)
    outfile.close()

    if network.link_combined:
        while True:
            try:
                outfile = open(segment_filepath, 'w', newline='',errors='ignore')
                break
            except PermissionError:
                print('segment.csv may be locked by other programs. please release it then press Enter to try again')
                input()
        writer = csv.writer(outfile)
        writer.writerow(['segment_id','link_id','ref_node_id','start_lr','end_lr','capacity','free_speed','bike_facility',
                         'ped_facility','parking','allowed_uses','l_lanes_added','r_lanes_added'])
        for segment in network.segment_list:
            line = [segment.segment_id, segment.link.link_id,segment.link.from_node.node_id,segment.start_lr,segment.end_lr,'',
                    '','','','','',segment.l_lanes_added,segment.r_lanes_added]
            writer.writerow(line)
        outfile.close()

    if network.movement_list:
        while True:
            try:
                outfile = open(movement_filepath, 'w', newline='',errors='ignore')     # ,errors='ignore'
                break
            except PermissionError:
                print('movement.csv may be locked by other programs. please release it then press Enter to try again')
                input()
        writer = csv.writer(outfile)
        writer.writerow(['mvmt_id', 'node_id', 'osm_node_id', 'name', 'ib_link_id','ib_lane','ob_link_id','ob_lane',
                         'lanes','ib_osm_node_id','ob_osm_node_id','type','penalty','capacity','ctrl_type','movement_str',
                         'geometry','volume','free_speed'])
        for mvmt in network.movement_list:
            node = mvmt.node
            from_node = mvmt.ib_link.from_node
            to_node = mvmt.ob_link.to_node
            if projection:
                geometry_ = geometry.LineString([from_node.geometry_xy, node.geometry_xy, to_node.geometry_xy])
            else:
                geometry_ = geometry.LineString([from_node.geometry, node.geometry, to_node.geometry])
            line = [mvmt.movement_id, node.node_id, node.osm_node_id, '', mvmt.ib_link.link_id, mvmt.ib_lane,
                    mvmt.ob_link.link_id, mvmt.ob_lane, mvmt.lanes, from_node.osm_node_id, to_node.osm_node_id,
                    '', '', '', '', mvmt.movement_str,geometry_]
            writer.writerow(line)

    if network.POI_list:
        while True:
            try:
                outfile = open(structure_filpath, 'w', newline='',errors='ignore')     # ,errors='ignore'
                break
            except PermissionError:
                print('POI.csv may be locked by other programs. please release it then press Enter to try again')
                input()
        writer = csv.writer(outfile)
        writer.writerow(['name', 'poi_id', 'osm_way_id', 'osm_relation_id','building','amenity','way','geometry','centroid',
                         'area', 'area_ft2'])
        for poi in network.POI_list:
            geometry_ = poi.geometry_xy if projection else poi.geometry
            centroid = poi.centroid_xy if projection else poi.centroid
            area = poi.geometry_xy.area
            area_ft2 = area * 10.7639
            line = [poi.name, poi.poi_id, poi.osm_way_id, poi.osm_relation_id, poi.building, poi.amenity, poi.way, geometry_,
                    centroid, round(area,1), round(area_ft2,1)]
            writer.writerow(line)


import setuptools

setuptools.setup(
    name='osm2gmns',
    version='0.3.2',
    author='Jiawei Lu, Xuesong Zhou',
    author_email='jiaweil9@asu.edu, xzhou74@asu.edu',
    url='https://github.com/jiawei92/OSM2GMNS',
    description='convert map data from OpenStreetMap to network files in GMNS format',
    long_description=open('README_pypi.rst').read(),
    license='GPLv3+',
    packages=['osm2gmns'],
    python_requires=">=3.6.0",
    install_requires=['pandas >= 0.24.0','shapely >= 1.7','protobuf >= 3.14.0'],
    classifiers=['License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
                 'Programming Language :: Python :: 3']
)
